# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE.
# Copyright (C) 2020 CERN.
#
# INSPIRE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE. If not, see <http://www.gnu.org/licenses/>.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.
from __future__ import absolute_import, division, print_function

import itertools
import json
from collections import Iterable

from json_merger.conflict import Conflict
from pyrsistent import thaw

from inspire_json_merger.utils import ORDER_KEY


def postprocess_results(merged, conflicts):
    """Run all postprocessing to provide output understandable by record-editor.

    Args:
        merged(dict): Merged document
        conflicts(list): List of all possible conflicts

    Returns: A tuple containing the resulted merged record in json format and a
        an list containing all generated conflicts.

    """

    conflicts, merged = postprocess_conflicts(conflicts, merged)
    conflicts_as_json = [json.loads(c.to_json()) for c in conflicts]
    flat_conflicts_as_json = remove_ordering_from_conflicts(
        list(itertools.chain.from_iterable(conflicts_as_json))
    )
    merged = remove_ordering_from_authors_merged(merged)

    return merged, flat_conflicts_as_json


def remove_ordering_from_conflicts(conflicts):
    """Cleans up ordering information in conflicts."""
    for conflict in conflicts:
        if isinstance(conflict["value"], dict):
            conflict["value"].pop(ORDER_KEY, None)
    return conflicts


def remove_ordering_from_authors_merged(merged):
    """Cleans up ordering information in merged record."""
    authors = []
    if "authors" in merged:
        for author in merged["authors"]:
            author = thaw(author)
            author.pop(ORDER_KEY, None)
            authors.append(author)
        merged["authors"] = authors
    return merged


def postprocess_conflicts(conflicts, merged):
    """Postprocessing conflicts to display only useful conflicts.

    Before flattening and serializing to JSON patch, MERGE conflict looks like this:
    ('MANUAL_MERGE', ('authors',), ( pmap(<ROOT>), pmap(<HEAD>), pmap(<UPDATE>) )),
    This function takes out all MANUAL_MERGE and do as follows:
        - Ignore ROOT
        - Update merged result with HEAD from conflict by inserting the author
            from HEAD at the right position.
        - Look for ADD_BACK_TO_HEAD conflict with identical content as HEAD
            change and remove it as it's already added to HEAD
    Args:
        conflicts(list): List of all possible conflicts
        merged(dict): Merged document

    Returns: A tuple containing the resulted merged record in json format and a
        an list containing all generated conflicts.
    """
    new_conflicts = []
    possible_duplicates = set()
    conflicts = sorted(conflicts, key=lambda conflict: conflict[0])
    # Sort by conflict type so we could process "ADD_BACK_TO_HEAD" after "MANUAL_MERGE"
    while conflicts:
        conflict = conflicts.pop()
        conflict_type, conflict_location, conflict_content = conflict
        if conflict_type == "MANUAL_MERGE" and conflict_location[0] == "authors":
            new_conflict, merged, head = _process_author_manual_merge_conflict(
                conflict, merged
            )
            if new_conflict:
                new_conflicts, conflicts = add_conflict(
                    new_conflict, new_conflicts, conflicts
                )
                possible_duplicates.add(head)
        elif not _is_conflict_duplicated(conflict, possible_duplicates):
            if conflict_type == "ADD_BACK_TO_HEAD":
                new_conflict, merged = _process_add_back_to_head(conflict, merged)
                new_conflicts, conflicts = add_conflict(
                    new_conflict, new_conflicts, conflicts
                )
            else:
                new_conflicts.append(conflict)
    return new_conflicts, merged


def add_conflict(conflict, processed_conflicts, unprocessed_conflicts):
    """Adds conflict which added something to `merged` dict so positions for other conflicts should be updated

    Args:
        conflict(Conflict): curently added conflict
        processed_conflicts(list): List of conflicts already added
        unprocessed_conflicts(list): List of conflicts which will be added
    """
    processed_conflicts = update_conflicts_list(conflict, processed_conflicts)
    unprocessed_conflicts = update_conflicts_list(conflict, unprocessed_conflicts)
    processed_conflicts.append(conflict)

    return processed_conflicts, unprocessed_conflicts


def update_conflicts_list(conflict, conflict_list):
    """Updates positions in provided conflict_list for conflict which will be added

    Args:
        conflict(Conflict): New conflict
        conflict_list(list): List of conflicts where positions should be updated
    """
    new_type, new_path, new_content = conflict
    for idx, processed_conflict in enumerate(conflict_list):
        path = processed_conflict[1]
        if path[0] == new_path[0] and len(path) > 1 and len(new_path) > 1 and path[1] >= new_path[1]:
            path = list(path)
            path[1] += 1
            path = tuple(path)
            conflict_list[idx] = Conflict(
                processed_conflict[0], path, processed_conflict[2]
            )
    return conflict_list


def _process_add_back_to_head(conflict, merged):
    """Process ADD_BACK_TO_HEAD conflicts differently than other conflicts.

    Replace all ADD_BACK_TO_HEAD conflicts to became REMOVE_FIELD conflicts
    also adds value from conflict back to merged_root
    REMOVE_FIELD conflict now points to proper element on list.
    """
    conflict_type, conflict_location, conflict_content = conflict
    if conflict_location[0] == "authors":
        position, merged["authors"] = _insert_to_list(
            conflict_content, merged["authors"]
        )
        insert_path = ("authors", position)
        new_conflict = Conflict("REMOVE_FIELD", insert_path, None)
        return new_conflict, merged
    else:
        insert_path, merged = _additem(conflict_content, merged, conflict_location)
        new_conflict = Conflict("REMOVE_FIELD", insert_path, None)
        return new_conflict, merged


def _additem(item, object, path):
    """Adds item to object on path.

    Args:
        item: Item to add
        object: List or Dictionary to which item should be added
        path(tuple): Path on which item should be added, every element of path is a separate element.
        Path represents place after which item should be added.

    Returns(tuple): Tuple containing path and item merged with item under proper path.
    """
    if path[0]:
        if path[0] == "-":
            position, object = _insert_to_list(item, object)
            return (position,), object
        else:
            try:
                idx = int(path[0])
                if len(path) == 1:
                    idx, object = _insert_to_list(item, object, idx)
                    return (idx,), object
                else:
                    new_path, object[idx] = _additem(item, object[idx], path[1:])
            except ValueError:
                if len(path) == 1:
                    if isinstance(item, list) or not isinstance(object[path[0]], list):
                        object[path[0]] = thaw(item)
                        return (path[0],), object
                    elif item not in object[path[0]]:
                        object[path[0]].append(thaw(item))
                        return (path[0], (len(object[path[0]]) - 1)), object
                else:
                    new_path, object[path[0]] = _additem(
                        item, object[path[0]], path[1:]
                    )
    return (path[0],) + new_path, object


def _is_conflict_duplicated(conflict, possible_duplicates):
    conflict_type, conflict_location, conflict_content = conflict
    if conflict_type == "ADD_BACK_TO_HEAD" and conflict_location[0] == "authors" and conflict_content in possible_duplicates:
        return True
    return False


def _process_author_manual_merge_conflict(conflict, merged):
    """Process author `MANUAL_MERGE` conflict.

    Conflict object is an tuple containing:
    (conflict_type, conflict_location, conflict_data)
    where `conflict_data` is a tuple of: (ROOT, HEAD, UPDATE).
    """
    _, _, (root, head, update) = conflict
    if head and head not in merged["authors"]:
        position, merged["authors"] = _insert_to_list(head, merged["authors"])
        new_conflict = Conflict("SET_FIELD", ("authors", position), update)
        return new_conflict, merged, head
    return None, merged, head


def _insert_to_list(item, objects_list, position=None):
    """Inserts value into list at proper position (as close to requested position as possible but not before it).

    If no position provided element will be added at the end of the list.
    Args:
        item: Value to insert into objects_list with `ORDER_KEY` key
        objects_list(list): List where value should be inserted
        position(int): If set then use this position instead `ORDER_KEY` key

    Returns(tuple): (position on which it was placed, merged objects_list).
    """
    item = thaw(item)
    if not position and ORDER_KEY in item:
        position = item[ORDER_KEY]
    if position is not None:
        for idx, element in enumerate(objects_list):
            if isinstance(element, Iterable) and ORDER_KEY in element:
                if element[ORDER_KEY] > position:
                    objects_list.insert(idx, item)
                    return idx, objects_list
            elif idx > position:
                objects_list.insert(idx, item)
                return idx, objects_list
    objects_list.append(item)
    return len(objects_list) - 1, objects_list



# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE.
# Copyright (C) 2014-2017 CERN.
#
# INSPIRE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE. If not, see <http://www.gnu.org/licenses/>.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""Pre-filters to transform the records before the merger operates."""

from __future__ import absolute_import, division, print_function

from functools import partial

import pyrsistent
from inspire_utils.record import get_value
from pyrsistent import freeze, thaw
from six.moves import zip
from inspire_json_merger.utils import ORDER_KEY


def remove_elements_with_source(source, field):
    """Remove all elements matching ``source`` in ``field``."""
    return freeze(
        [element for element in field if element.get('source', '').lower() != source]
    )


def keep_only_update_source_in_field(field, root, head, update):
    """Remove elements from root and head where ``source`` matches the update.

    This is useful if the update needs to overwrite all elements with the same
    source.

    .. note::
        If the update doesn't contain exactly one source in ``field``, the
        records are returned with no modifications.

    Args:
        field (str): the field to filter out.
        root (pmap): the root record, whose ``field`` will be cleaned.
        head (pmap): the head record, whose ``field`` will be cleaned.
        update (pmap): the update record, from which the ``source`` is read.

    Returns:
        tuple: ``(root, head, update)`` with some elements filtered out from
            ``root`` and ``head``.
    """
    update_thawed = thaw(update)
    update_sources = {source.lower() for source in get_value(update_thawed, '.'.join([field, 'source']), [])}
    if not update_sources:
        # If there is no field or source then fallback for source to `aquisition_source.source`
        source = get_value(update_thawed, "acquisition_source.source")
        if source:
            update_sources = {source.lower()}
    if len(update_sources) != 1:
        return root, head, update
    source = update_sources.pop()

    if field in root:
        root = root.set(field, remove_elements_with_source(source, root[field]))
    if field in head:
        head = head.set(field, remove_elements_with_source(source, head[field]))

    return root, head, update


def filter_curated_references(root, head, update):
    """Remove references from either ``head`` or ``update`` depending on curation.

    If references have been curated, then it removes all references from the
    update to keep the existing ones. Otherwise, it removes all references from
    the head to force replacement with the update ones.

    Args:
        root (pmap): the root record.
        head (pmap): the head record.
        update (pmap): the update record.

    Returns:
        tuple: ``(root, head, update)`` with ``references`` removed from ``root`` and either
        ``head`` or ``update``.
    """
    if 'references' not in head or 'references' not in update:
        return root, head, update

    references_curated = are_references_curated(root.get('references', []), head.get('references', []))
    if 'references' in root:
        root = root.remove('references')
    if references_curated:
        update = update.remove('references')
    else:
        head = head.remove('references')

    return root, head, update


def filter_publisher_references(root, head, update):
    """Remove references from ``update`` if there are any in ``head``.

    This is useful when merging a record from a publisher with an update form arXiv,
    as arXiv should never overwrite references from the publisher.
    """
    if 'references' in head:
        root = _remove_if_present(root, 'references')
        update = _remove_if_present(update, 'references')

    return root, head, update


def update_authors_with_ordering_info(root, head, update):
    """Adds ordering information into authors entries.

    This is helpful so we won't loose correct ordering on conflicts.
    Args:
        root (pmap): the root record.
        head (pmap): the head record.
        update (pmap): the update record.

    Returns: A tuple containing root, head and update with authors data
        enriched with ordering information

    """
    if 'authors' in head:
        head = head.update({"authors": _update_authors_list_with_ordering_data(head["authors"])})
    return root, head, update


def _update_authors_list_with_ordering_data(input_list):
    positions = iter(range(len(input_list)))
    return input_list.transform([pyrsistent.ny], lambda element: element.set(ORDER_KEY, (next(positions))))


def are_references_curated(root_refs, head_refs):
    if not root_refs:
        return any('legacy_curated' in head_ref for head_ref in head_refs)

    if len(root_refs) != len(head_refs):
        return True

    if all(ref_almost_equal(root, head) for (root, head) in zip(root_refs, head_refs)):
        return False

    return True


def ref_almost_equal(root_ref, head_ref):
    return _normalize_ref(root_ref) == _normalize_ref(head_ref)


def _normalize_ref(ref):
    ref = _remove_if_present(ref, 'record')
    ref = _remove_if_falsy(ref, 'curated_relation')
    ref = _remove_if_present(ref, 'raw_refs')
    ref = ref.transform(['reference'], lambda reference: _remove_if_present(reference, 'misc'))
    ref = ref.transform(['reference'], lambda reference: _remove_if_present(reference, 'authors'))
    return ref


def _remove_if_present(pmap, key):
    try:
        return pmap.remove(key)
    except KeyError:
        return pmap


def _remove_if_falsy(pmap, key):
    try:
        if not pmap[key]:
            return pmap.remove(key)
        return pmap
    except KeyError:
        return pmap


def remove_duplicated_titles(root, head, update):
    new_elements = []
    for data in (root, head, update):
        titles = data.get('titles')
        if not titles:
            new_elements.append(data)
            continue
        titles_dict = dict()
        to_delete = []
        for title in titles:
            title_data = (title['title'], title.get('subtitle'))
            if title_data not in titles_dict:
                titles_dict[title_data] = title
            else:
                if title['source'].lower() != 'arxiv':
                    to_delete.append(titles_dict[title_data])
                    titles_dict[title_data] = title
                else:
                    to_delete.append(title)
        for object_to_remove in to_delete:
            titles = titles.remove(object_to_remove)
        new_elements.append(data.update({'titles': titles}))
    print(new_elements)
    return new_elements


filter_documents_same_source = partial(keep_only_update_source_in_field, 'documents')
filter_figures_same_source = partial(keep_only_update_source_in_field, 'figures')



# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE.
# Copyright (C) 2014-2017 CERN.
#
# INSPIRE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE. If not, see <http://www.gnu.org/licenses/>.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""Helper functions for authors."""

from __future__ import absolute_import, division, print_function

import re
import six
from pyrsistent import freeze, thaw
from six.moves import zip

split_on_re = re.compile(r'[\.\s-]')

ORDER_KEY = "__pos"


def scan_author_string_for_phrases(s):
    """Scan a name string and output an object representing its structure.
    Example:
        Sample output for the name 'Jingleheimer Schmitt, John Jacob, XVI.' is::
            {
                'TOKEN_TAG_LIST' : ['lastnames', 'nonlastnames', 'titles', 'raw'],
                'lastnames'      : ['Jingleheimer', 'Schmitt'],
                'nonlastnames'   : ['John', 'Jacob'],
                'titles'         : ['XVI.'],
                'raw'            : 'Jingleheimer Schmitt, John Jacob, XVI.'
            }
    :param s: the input to be lexically tagged
    :type s: string
    :returns: dict of lexically tagged input items.
    :rtype: dict
    """

    if not isinstance(s, six.text_type):
        s = s.decode('utf-8')

    retval = {
        'TOKEN_TAG_LIST': [
            'lastnames',
            'nonlastnames',
            'titles',
            'raw'],
        'lastnames': [],
        'nonlastnames': [],
        'titles': [],
        'raw': s}
    l = s.split(',')  # noqa: E741
    if len(l) < 2:
        # No commas means a simple name
        new = s.strip()
        new = new.split(' ')
        if len(new) == 1:
            retval['lastnames'] = new        # rare single-name case
        else:
            retval['lastnames'] = new[-1:]
            retval['nonlastnames'] = new[:-1]
            for tag in ['lastnames', 'nonlastnames']:
                retval[tag] = [x.strip() for x in retval[tag]]
                retval[tag] = [re.split(split_on_re, x)
                               for x in retval[tag]]
                # flatten sublists
                retval[tag] = [item for sublist in retval[tag]
                               for item in sublist]
                retval[tag] = [x for x in retval[tag] if x != '']
    else:
        # Handle lastname-first multiple-names case
        retval['titles'] = l[2:]             # no titles? no problem
        retval['nonlastnames'] = l[1]
        retval['lastnames'] = l[0]
        for tag in ['lastnames', 'nonlastnames']:
            retval[tag] = retval[tag].strip()
            retval[tag] = re.split(split_on_re, retval[tag])
            # filter empty strings
            retval[tag] = [x for x in retval[tag] if x != '']
        retval['titles'] = [x.strip() for x in retval['titles'] if x != '']

    return retval


def filter_conflicts(conflicts_list, fields):
    """Use this function to automatically filter all the entries defined for a
    given rule.

    Params:
        conflicts_list(List[Conflict]): the list of conflicts to filter.
        fields(List[str]): fields to filter out, using an accessor syntax of
            the form ``field.subfield.subsubfield``.

    Return:
        List[Conflict]: the given list filtered by `fields`
    """
    for field in fields:
        conflicts_list = filter_conflicts_by_path(conflicts_list, field)

    return conflicts_list


def filter_conflicts_by_path(conflict_list, to_delete_path):
    """Filter a list of conflict for the given string. The string represents
    the path of the conflict in the form ``field.subfield.subsubfield``.

    Example:
        conflict_list = [
            ('SET_FIELD', ('figures', 0, 'key'), 'figure1.png'),
            ('SET_FIELD', ('figures', 1, 'key'), 'figure2.png')
        ]

        to_delete_path = 'figures.keys'
    """
    return [conf for conf in conflict_list if not is_to_delete(conf, to_delete_path)]


def is_to_delete(conflict, keys_path):
    to_delete = keys_path.split('.')
    conflict_path = conflict_to_list(conflict)
    if conflict[0] == 'MANUAL_MERGE':
        return False
    if len(to_delete) > len(conflict_path):
        return False

    return all(x == y for (x, y) in zip(to_delete, conflict_path))


def conflict_to_list(conflict):
    path = conflict[1]
    return [p for p in path if not isinstance(p, int)]


def filter_records(root, head, update, filters=()):
    """Apply the filters to the records."""
    root, head, update = freeze(root), freeze(head), freeze(update)
    for filter_ in filters:
        root, head, update = filter_(root, head, update)

    return thaw(root), thaw(head), thaw(update)
